import asyncio
import dataclasses
import docopt
import json
import os.path
import sys
from typing import Any, Mapping, Optional

from . import __version__, Client, LocalSpecification, load_notebook_models
from .modeling import Model

_COMMAND = "python -m opvious"

_DOC = f"""Opvious Python CLI

Command line utilities for interacting with the Opvious API from Python. The
underlying client can be configured using the `OPVIOUS_ENDPOINT` and
`OPVIOUS_TOKEN` environment variables. Refer to the standard CLI
(https://www.npmjs.com/package/opvious-cli) for additional operations.

Usage:
    {_COMMAND} register-notebook PATH [MODEL]
        [-dn NAME] [-t TAGS] [--allow-empty]
    {_COMMAND} register-sources GLOB [-dn NAME] [-t TAGS]
    {_COMMAND} export-notebook-model PATH [MODEL] [-a PATH]
    {_COMMAND} (-h | --help)
    {_COMMAND} --version

Options:
    --allow-empty             Do not throw an error if no models were found in
                              a notebook. Requires `--dry-run` to be set.
    -a, --assembly-path PATH  Path where to store the exported model. Defaults
                              to the model's name with a `.proto` extension.
    -d, --dry-run             Validate the specification but do not store it on
                              the server. When this option is enabled,
                              notebooks can have more than one model.
    -h, --help                Show this message.
    -n, --name NAME           Formulation name. By default this name is
                              inferred from the file's name, omitting the
                              extension.
    -t, --tags TAGS           Comma-separated list of tags. By default only the
                              `latest` tag is added.
    --version                 Show SDK version.
"""


class _SpecificationHandler:
    def __init__(
        self, client: Client, tags: Optional[str] = None, dry_run=False
    ) -> None:
        self._client = client
        self._tags = tags.split() if tags else None
        self._dry_run = dry_run

    async def _handle(self, spec: LocalSpecification, name: str) -> None:
        if self._dry_run:
            annotated = await self._client.annotate_specification(spec)
            ann = annotated.annotation
            assert ann is not None
            if ann.issue_count > 0:
                raise Exception(
                    f"{ann.issue_count} issue(s) in specification: "
                    + json.dumps(
                        [
                            dataclasses.asdict(i)
                            for issues in ann.issues.values()
                            for i in issues
                        ]
                    )
                )
            print("Specification is valid")
        else:
            await self._client.register_specification(
                spec,
                formulation_name=name,
                tag_names=self._tags,
            )
            print(f"Registered specification in formulation {name}")

    async def handle_notebook(
        self,
        path: str,
        model_name: Optional[str],
        name: Optional[str],
        allow_empty: bool,
    ) -> None:
        models = _load_notebook_models(path, model_name)
        if self._dry_run:
            return
        _name, model = _singleton_model(models)
        await self._handle(model.specification(), name or _default_name(path))

    async def handle_sources(self, glob: str, name: Optional[str]) -> None:
        if name is None:
            name = _default_name(glob)
        spec = LocalSpecification.globs(glob)
        await self._handle(spec, name)


def _default_name(path: str) -> str:
    return os.path.splitext(os.path.basename(path))[0]


def _load_notebook_models(
    path: str,
    model_name: Optional[str],
) -> dict[str, Model]:
    sn = load_notebook_models(path, allow_empty=True)
    if model_name is None:
        return {k: v for k, v in sn.__dict__.items() if isinstance(v, Model)}
    return {model_name: getattr(sn, model_name)}


def _singleton_model(models: dict[str, Model]) -> tuple[str, Model]:
    if len(models) != 1:
        raise Exception(
            "Notebook has 0 or 2+ models, please specify a model "
            "name to select one"
        )
    return next(iter(models.items()))


async def _export_notebook_model(
    client: Client,
    notebook_path: str,
    model_name: Optional[str] = None,
    export_path: Optional[str] = None,
) -> None:
    # TODO: Support transformations by accepting an additional variable name.
    models = _load_notebook_models(notebook_path, model_name)
    name, model = _singleton_model(models)
    if not export_path:
        export_path = f"{name}.proto"
    with open(export_path, "bw+") as writer:
        await client.export_specification(model.specification(), writer)


async def _run(args: Mapping[str, Any]) -> None:
    client = Client.from_environment()
    if not client:
        raise Exception("Missing OPVIOUS_ENDPOINT environment variable")

    if args["export-notebook-model"]:
        await _export_notebook_model(
            client,
            notebook_path=args["PATH"],
            model_name=args["MODEL"],
            export_path=args["--assembly-path"],
        )
        return

    handler = _SpecificationHandler(client, args["--tags"], args["--dry-run"])
    if args["register-notebook"]:
        await handler.handle_notebook(
            args["PATH"],
            model_name=args["MODEL"],
            name=args["--name"],
            allow_empty=args["--allow-empty"],
        )
    elif args["register-sources"]:
        await handler.handle_sources(args["GLOB"], name=args["--name"])
    else:
        raise Exception("Unexpected command")


if __name__ == "__main__":
    argv = _COMMAND.split()[1:] + sys.argv[1:]
    args = docopt.docopt(_DOC, argv=argv, version=__version__)
    asyncio.run(_run(args))



from __future__ import annotations

import dataclasses
from datetime import datetime, timedelta
from typing import Any, Optional, cast

from ..common import (
    Annotation,
    Json,
    Uuid,
    decode_datetime,
    decode_annotations,
    if_present,
)
from .outcomes import (
    SolveOutcome,
    failed_outcome_from_graphql,
    solve_outcome_from_graphql,
)
from .solves import ProblemSummary, problem_summary_from_json
from .tensors import Value


@dataclasses.dataclass(frozen=True)
class QueuedSolve:
    """Queued optimization attempt

    Solves are queued via :meth:`.Client.queue_solve`, existing queued solves
    can be retrieved from their UUID via :meth:`.Client.fetch_solve`.
    """

    uuid: Uuid
    """The solve's unique identifier"""

    annotations: list[Annotation]
    """Annotation metadata"""

    outcome: Optional[SolveOutcome]
    """Final solve outcome, if available"""

    enqueued_at: datetime
    """The time the solve was created"""

    dequeued_at: Optional[datetime]
    """The time the solve started running"""

    completed_at: Optional[datetime]
    """The time the solve completed"""

    problem_summary: Optional[ProblemSummary] = dataclasses.field(repr=False)
    """Summary information about the solved problem"""

    options: Json = dataclasses.field(repr=False)
    transformations: Json = dataclasses.field(repr=False)
    strategy: Json = dataclasses.field(repr=False)

    @property
    def duration(self) -> Optional[timedelta]:
        """The solve's runtime, if it is complete"""
        return (
            self.completed_at - self.enqueued_at if self.completed_at else None
        )


def queued_solve_from_graphql(
    data: Json, attempt_data: Optional[Json] = None
) -> QueuedSolve:
    attempt_data = attempt_data or data["attempt"]

    if data["failure"]:
        outcome = cast(
            SolveOutcome, failed_outcome_from_graphql(data["failure"])
        )
    elif data["outcome"]:
        outcome = solve_outcome_from_graphql(data["outcome"])
    else:
        outcome = None

    return QueuedSolve(
        uuid=data["uuid"],
        enqueued_at=decode_datetime(attempt_data["startedAt"]),
        dequeued_at=if_present(data["dequeuedAt"], decode_datetime),
        completed_at=if_present(attempt_data["endedAt"], decode_datetime),
        annotations=decode_annotations(attempt_data["annotations"]),
        options=data["options"],
        transformations=data["transformations"],
        strategy=data["strategy"],
        outcome=outcome,
        problem_summary=if_present(
            data["problemSummary"], problem_summary_from_json
        ),
    )


@dataclasses.dataclass(frozen=True)
class SolveNotification:
    """Solve progress update notification"""

    dequeued: bool
    """Whether the solve has already been dequeued"""

    relative_gap: Optional[Value]
    """The latest relative gap"""

    lp_iteration_count: Optional[int]
    """The latest LP iteration count"""

    cut_count: Optional[int]
    """The latest cut count"""


def solve_notification_from_graphql(
    dequeued: bool, data: Any = None
) -> SolveNotification:
    return SolveNotification(
        dequeued=dequeued,
        relative_gap=data["relativeGap"] if data else None,
        lp_iteration_count=data["lpIterationCount"] if data else None,
        cut_count=data["cutCount"] if data else None,
    )
